from typing import List, Literal, Union

from typing_extensions import TypedDict

DataSource = Literal["issues", "boards"]

JiraFieldName = str


class SelectFieldDefinition(TypedDict):
    expression: str
    column: str


Field = Union[SelectFieldDefinition, JiraFieldName]

JQLString = str

CustomFilterFieldName = str

JqlList = List[str]

ExpressionList = List[str]

# We have to use the alternative method of defining the TypedDict since
# one of our dictionary fields' names is a reserved word
QueryDefinition = TypedDict(
    "QueryDefinition",
    {
        "select": List[Field],
        "from": DataSource,
        "where": JqlList,
        "having": ExpressionList,
        "expand": List[JiraFieldName],
        # "having": Dict[CustomFilterFieldName, Any],
    },
    total=False,
)


class ViewerDefinitionDict(TypedDict, total=False):
    csv: str


class ConfigDict(TypedDict, total=False):
    instance_url: str
    username: str
    password: str
    viewers: ViewerDefinitionDict



import argparse

from rich.table import Table

from ..plugin import BaseCommand


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return "Searches Jira for fields matching your search query."

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--search-term",
            help="A partial string to match against your search field's name or key.",
        )

    def handle(self) -> None:
        table = Table(title="Matching Fields")
        table.add_column(header="key", style="cyan")
        table.add_column(header="type", style="green")
        table.add_column(header="name", style="bright_cyan")

        search_term = ""
        if self.options.search_term:
            search_term = self.options.search_term.lower()

        fields = sorted(self.jira.fields(), key=lambda field: field["name"])
        for field in fields:
            if (
                search_term not in field["key"].lower()
                and search_term not in field["name"].lower()
            ):
                continue

            table.add_row(
                field["key"], field.get("schema", {}).get("type", ""), field["name"]
            )

        self.console.print(table)



from __future__ import annotations

import argparse
import logging
from abc import ABCMeta, abstractmethod
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, Optional, Type, cast

import pkg_resources

import keyring
from jira import JIRA
from rich.console import Console

from .constants import APP_NAME
from .exceptions import ConfigurationError
from .types import ConfigDict
from .utils import save_config

if TYPE_CHECKING:
    from .query import Query

logger = logging.getLogger(__name__)


def get_installed_commands() -> Dict[str, Type[BaseCommand]]:
    possible_commands: Dict[str, Type[BaseCommand]] = {}
    for entry_point in pkg_resources.iter_entry_points(group="jira_select.commands"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, BaseCommand):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `jira_select.plugin.BaseCommand`.",
                entry_point,
            )
            continue
        possible_commands[entry_point.name] = loaded_class

    return possible_commands


class BaseCommand(metaclass=ABCMeta):
    _jira: Optional[JIRA] = None

    def __init__(self, config: ConfigDict, options: argparse.Namespace):
        self._config: ConfigDict = config
        self._options: argparse.Namespace = options
        self._console = Console()
        super().__init__()

    @property
    def config(self) -> ConfigDict:
        return self._config

    def save_config(self) -> None:
        save_config(self.config, self.options.config)

    @property
    def options(self) -> argparse.Namespace:
        return self._options

    @property
    def console(self) -> Console:
        return self._console

    @property
    def jira(self) -> JIRA:
        if self._jira is None:
            instance_url = self.config.get("instance_url") or self.options.instance_url
            if not instance_url:
                raise ConfigurationError("instance_url not set")

            username = self.config.get("username") or self.options.username
            if not username:
                raise ConfigurationError("username not set")

            password = self.config.get("password") or self.options.password
            if not password:
                password = keyring.get_password(APP_NAME, instance_url + username)
                if not password:
                    raise ConfigurationError(
                        f"Password not stored for {instance_url} user {username}; "
                        f"use the 'store-password' command to store the password "
                        f"for this user account in your system keyring."
                    )
            self._jira = JIRA(
                server=instance_url,
                options={"agile_rest_path": "agile",},
                basic_auth=(username, password),
            )

        return self._jira

    @classmethod
    def get_help(cls) -> str:
        return ""

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        pass

    @abstractmethod
    def handle(self) -> None:
        ...


def get_installed_formatters() -> Dict[str, Type[BaseFormatter]]:
    possible_formatters: Dict[str, Type[BaseFormatter]] = {}
    for entry_point in pkg_resources.iter_entry_points(group="jira_select.formatters"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, BaseFormatter):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `jira_select.plugin.BaseFormatter`.",
                entry_point,
            )
            continue
        possible_formatters[entry_point.name] = loaded_class

    return possible_formatters


class BaseFormatter(metaclass=ABCMeta):
    def __init__(self, query: Query, stream: IO[str]):
        self._query = query
        self._stream = stream

    @classmethod
    @abstractmethod
    def get_file_extension(cls) -> str:
        ...

    @property
    def query(self):
        return self._query

    @property
    def stream(self):
        return self._stream

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        # Exception handling here
        self.close()

    def open(self):
        return

    def close(self):
        return

    @abstractmethod
    def writerow(self, row: Dict[str, Any]):
        ...


def get_installed_functions(jira: JIRA = None) -> Dict[str, Callable]:
    possible_commands: Dict[str, BaseFunction] = {}
    for entry_point in pkg_resources.iter_entry_points(group="jira_select.functions"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, BaseFunction):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `jira_select.plugin.BaseFunction`.",
                entry_point,
            )
            continue
        possible_commands[entry_point.name] = loaded_class(jira)

    return cast(Dict[str, Callable], possible_commands)


class BaseFunction(metaclass=ABCMeta):
    def __init__(self, jira: Optional[JIRA]):
        self._jira = jira

    @property
    def jira(self):
        assert self._jira

        return self._jira

    @abstractmethod
    def process(self, *args, **kwargs) -> Optional[Any]:
        ...

    def __call__(self, *args, **kwargs) -> Optional[Any]:
        return self.process(*args, **kwargs)



import argparse

from rich.traceback import install as enable_rich_traceback

from .plugin import get_installed_commands
from .utils import get_config


def main():
    enable_rich_traceback()
    commands = get_installed_commands()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        "-c",
        default=None,
        type=str,
        help="Path to configuration file to load.",
    )
    parser.add_argument(
        "--instance-url",
        "-i",
        default=None,
        type=str,
        help="URL of the Jira instance to connect to (if you cannot use 'configure')",
    )
    parser.add_argument(
        "--username",
        "-u",
        default=None,
        type=str,
        help="Username to use for connecting to Jira (if you cannot use 'configure')",
    )
    parser.add_argument(
        "--password",
        "-p",
        default=None,
        type=str,
        help=(
            "Password to use for connecting to Jira.  NOT RECOMMENDED: use the "
            "'store-password' or 'configure' command instead."
        ),
    )
    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    for cmd_name, cmd_class in commands.items():
        parser_kwargs = {}

        cmd_help = cmd_class.get_help()
        if cmd_help:
            parser_kwargs["help"] = cmd_help

        subparser = subparsers.add_parser(cmd_name, **parser_kwargs)
        cmd_class.add_arguments(subparser)

    args = parser.parse_args()

    config = get_config(path=args.config)

    command = commands[args.command](config=config, options=args)
    command.handle()



from typing import Any, Callable, Dict, Generator, List, Optional

from jira import JIRA, Issue

from .exceptions import UserError
from .plugin import get_installed_functions
from .types import QueryDefinition, SelectFieldDefinition
from .utils import get_field_data, parse_select_definition


class Query:
    def __init__(self, jira: JIRA, definition: QueryDefinition, emit_omissions=False):
        self._definition: QueryDefinition = definition
        self._jira: JIRA = jira
        self._functions: Dict[str, Callable] = get_installed_functions(jira)
        self._emit_omissions: bool = emit_omissions

    @property
    def jira(self) -> JIRA:
        return self._jira

    @property
    def definition(self) -> QueryDefinition:
        return self._definition

    def get_fields(self) -> List[SelectFieldDefinition]:
        fields: List[SelectFieldDefinition] = []

        for field in self.definition["select"]:
            fields.append(parse_select_definition(field))

        return fields

    def _generate_row_dict(self, row: Any) -> Dict[str, Any]:
        result: Dict[str, Any] = {}

        for field_defn in self.get_fields():
            result[field_defn["column"]] = get_field_data(
                row, field_defn["expression"], functions=self._functions
            )

        return result

    def _get_jql(self) -> str:
        where = self.definition.get("where")

        if where is None:
            return ""
        elif isinstance(where, List):
            return " AND ".join(self.definition["where"])

        raise UserError(f"Could not generate JQL from {where}.")

    def _get_issues(self) -> Generator[Issue, None, None]:
        jql = self._get_jql()

        expand = self.definition.get("expand", [])

        startAt = 0
        maxResults = float("inf")
        while startAt < maxResults:
            results = self.jira.search_issues(jql, startAt=startAt)
            maxResults = results.total
            for result in results:
                yield self.jira.issue(result.key, expand=",".join(expand))
                startAt += 1

    def _get_issues_having(self) -> Generator[Optional[Issue], None, None]:
        for row in self._get_issues():
            include_row = True
            for having in self.definition.get("having", []):
                if not get_field_data(row, having, self._functions):
                    include_row = False
                    break

            if include_row:
                yield row
            elif self._emit_omissions:
                yield None

    def _get_iterator(self) -> Generator[Optional[Any], None, None]:
        source = self.definition["from"]
        source = self.definition["from"]

        if source == "issues":
            yield from self._get_issues_having()
        else:
            raise NotImplementedError(f"No search for source {source} implemented.")

    def count(self) -> int:
        jql = self._get_jql()

        return self.jira.search_issues(jql).total

    def __iter__(self) -> Generator[Optional[Dict[str, Any]], None, None]:
        for row in self._get_iterator():
            if row is None:
                yield None
            else:
                yield self._generate_row_dict(row)



import logging
import os
import re
from typing import Any, Callable, Dict, Optional, Union

from appdirs import user_config_dir
from simpleeval import AttributeDoesNotExist, simple_eval
from yaml import safe_dump, safe_load

from .constants import APP_NAME
from .types import ConfigDict, SelectFieldDefinition

FIELD_DISPLAY_DEFN_RE = re.compile(r'^(?P<expression>.*) as "(?P<column>.*)"$')


logger = logging.getLogger(__name__)


def get_default_config_path() -> str:
    root_path = user_config_dir(APP_NAME, "coddingtonbear")
    os.makedirs(root_path, exist_ok=True)
    return os.path.join(root_path, "config.yaml",)


def get_config(path: str = None) -> ConfigDict:
    if path is None:
        path = get_default_config_path()

    if not os.path.isfile(path):
        return {}

    with open(path, "r") as inf:
        return safe_load(inf)


def save_config(data: ConfigDict, path: str = None) -> None:
    if path is None:
        path = get_default_config_path()

    with open(path, "w") as outf:
        safe_dump(data, outf)


def parse_select_definition(
    expression: Union[str, SelectFieldDefinition]
) -> SelectFieldDefinition:
    if isinstance(expression, str):
        as_match = FIELD_DISPLAY_DEFN_RE.match(expression)
        column = expression
        expression = expression
        if as_match:
            match_dict = as_match.groupdict()
            expression = match_dict["expression"]
            column = match_dict["column"]

        return {"expression": expression, "column": column}
    return expression


def get_field_data(
    row: Any, expression: str, functions: Optional[Dict[str, Callable]] = None
) -> Any:
    if functions is None:
        functions = {}

    names: Dict[str, Any] = {}
    if hasattr(row, "key"):
        names["key"] = row.key

    if hasattr(row, "fields"):
        for field_name in dir(row.fields):
            names[field_name] = getattr(row.fields, field_name)

    try:
        return simple_eval(expression, names=names, functions=functions)
    except AttributeDoesNotExist:
        return None



from typing import Any, List, Optional

from ..plugin import BaseFunction


class Function(BaseFunction):
    def process(self, array: Optional[List[Any]]) -> Optional[int]:  # type: ignore[override]
        if array is None:
            return None

        try:
            return len(array)
        except IndexError:
            return None



from typing import Any, Optional

from ..plugin import BaseFunction


class Function(BaseFunction):
    def process(self, *options) -> Optional[Any]:  # type: ignore[override]
        for option in options:
            if option is not None:
                return option

        return None
